import 'dart:async';

import 'build.dart';

/// A generator for a `packages.bzl` macro file.
class BazelMacroFile {
  final String _packageName;
  final List<NewLocalRepository> _pubDeps;

  /// Creates a new [BazelMacroFile] from [packages].
  ///
  /// Uses [resolvePath] to determine the local path for a given package.
  factory BazelMacroFile.fromPackages(
    String workspacePackage,
    Iterable<String> packages,
    String resolvePath(String packageName),
  ) {
    final repos = packages.map/*<NewLocalRepository>*/((d) {
      return new NewLocalRepository(
        name: d,
        path: resolvePath(d),
        buildFile: '.bazelify/$d.BUILD',
      );
    });
    return new BazelMacroFile.fromRepositories(
      workspacePackage,
      new List<NewLocalRepository>.unmodifiable(repos),
    );
  }

  /// Creates a new [BazelMacroFile] from the package name and repositories.
  BazelMacroFile.fromRepositories(this._packageName, this._pubDeps);

  @override
  String toString() {
    var buffer = new StringBuffer(
        '# Automatically generated by "pub global run bazel:bazelify".\n'
        '# DO NOT MODIFY BY HAND\n\n'
        'PUB_PACKAGE_NAME = "$_packageName"\n\n'
        '# Generated automatically for package:$_packageName|pubspec.yaml\n'
        'def bazelify():\n');
    _pubDeps.map/*<String>*/((d) {
      var macro = d.toString().split('\n').map((r) => '    $r').toList();
      macro = macro.take(macro.length - 1);
      return macro.join('\n');
    }).forEach(buffer.writeln);
    return buffer.toString();
  }
}

/// Turns a local directory into a `new_local_repository` Bazel repository.
///
/// See: https://www.bazel.io/versions/master/docs/be/workspace.html#new_local_repository
class NewLocalRepository {
  /// Name of the package.
  final String name;

  /// Local file path of the package.
  final String path;

  /// Where the generated BUILD file should be.
  final String buildFile;

  /// Create a `new_local_repository` macro.
  NewLocalRepository({
    this.name,
    this.path,
    this.buildFile,
  });

  /// Returns the contents of a suitable BUILD file for the repository.
  Future<BuildFile> resolveBuild() => BuildFile.fromPackageDir(path);

  @override
  String toString() {
    var buffer = new StringBuffer()
      ..writeln('native.new_local_repository(')
      ..writeln('    name = "$name",')
      ..writeln('    path = "$path",')
      ..writeln('    build_file = ".bazelify/$name.BUILD",')
      ..writeln(')');
    return buffer.toString();
  }
}
